module Error(putLogLn,attempt, openLog, closeLog, tryElse, tryMost, tryMost_, catchMost, tryMap, tryMapM, first, handleMost, ioElse, withErrorLog) where

import Control.Exception as E
import IO
import IOExts
import Monad
import Control.Concurrent.MVar

------------
-- Error log
------------

{-# NOINLINE ior #-}
ior :: MVar Handle
ior = unsafePerformIO $ newMVar stderr

withErrorLog :: String -> IO a -> IO a
withErrorLog fn action = E.bracket (openFile fn AppendMode) hClose $ \h -> do
	hSetBuffering h LineBuffering
	oh <- swapMVar ior h
	v <- action
	swapMVar ior oh
	return v
    

openLog :: String -> IO ()
openLog fn = handle (const $ return ()) $ do
	h <- openFile fn AppendMode
	hSetBuffering h NoBuffering
	swapMVar ior h
	return ()

closeLog :: IO ()
closeLog = do
   h <- takeMVar ior  
   putMVar ior stderr
   when (h /= stderr) $ hClose h

logBracket = E.bracket (takeMVar ior) (putMVar ior)


putLogLn :: String -> IO ()
putLogLn s = logBracket (\h -> hPutStrLn h s)


attempt :: IO a -> IO ()
attempt action = tryMost action >>= \x -> case x of 
    Left e -> putLogLn ("Exception Caught: " ++ show e)
    Right _ -> return ()


tryElse r x = tryMost x >>= \y -> case y of
    Left x -> (putLogLn ("tryElse ExceptionCaught:" ++ show x)) >> return r
    Right v -> return v

tryMost_ x = tryMost x >> return ()

tryMap :: (a -> b) -> [a] -> IO [b]
tryMap f xs = do 
    ys <- mapM (tryMost . evaluate . f ) xs
    return [y|(Right y) <- ys]
    
tryMapM :: (a -> IO b) -> [a] -> IO [b]
tryMapM f xs = do 
    ys <- mapM (tryMost . f ) xs
    return [y|(Right y) <- ys]

tryMost = E.tryJust passKilled

passKilled (AsyncException ThreadKilled) = Nothing
passKilled x = Just x

catchMost = E.catchJust passKilled
handleMost = E.handleJust passKilled

first :: [IO a] -> IO a
first [] = fail "empty  argument to first"
first [x] = x
first (x:xs) = tryMost x >>= \z -> case z of
    Left _ -> first xs
    Right v -> return v

ioElse :: IO a -> IO a -> IO a
ioElse a b = tryMost a >>= \x -> case x of 
    Left _ -> b
    Right x -> return x
