module Gale(GaleContext,Puff(..),galeNextPuff,showPuff, newGaleContext, Fragment(..), Category, catParseNew, 
    catShowNew, catShowOld, subCategory, connectGaleContext, reconnectGaleContext, closeGaleContext,
    getFragmentString, getFragmentStrings, writePuffs, readPuffs, Signature(..),galeInit, Key(..)) where


import Bits
import Char(chr,ord)
import Concurrent
import Control.Exception
import Directory
import Int
import IO
import IOExts
import List
import Monad(when)
import Network
import System
import Time
import Word

import ConcurrentUtil
import Error
import GenUtil
import Puff
import RSA
import SimpleParser
import EIO


-- TODO - prove concurrent-correctness, make sure all network errors are accounted for.

-------------------
-- Gale Constants
-------------------

galePort :: PortNumber
galePort = 11512
hostStrings s = [s, "gale." ++ s, s ++ ".gale.org."]

galeRSAModulusBits = 1024
galeRSAModulusLen = (galeRSAModulusBits + 7) `div` 8
galeRSAPrimeBits = (galeRSAModulusBits + 1) `div` 2
galeRSAPrimeLen = (galeRSAPrimeBits + 7) `div` 8
--galeEncryptedKeyLen = galeRSAModulusLen
--galeSignatureLen = galeRSAModulusLen

data GaleContext = GaleContext { channel :: Chan Puff, proxy :: MVar [String], categories :: MVar [Category], threads :: MVar [ThreadId] }


subCategory :: Category -> Category -> Bool
(x1,y1) `subCategory` (x2,y2) = x2 `isPrefixOf` x1 && y2 `isSuffixOf` y1


newGaleContext gps cs = do
    cs' <- tryMap catParseNew cs
    c <- newChan 
    mv <- newEmptyMVar
    cmv <- newMVar cs'
    pmv <- newMVar gps
    let gc = GaleContext { channel = c, categories = cmv, threads = mv, proxy = pmv}
    connectGaleContext gc
    return gc

spc [] = []
spc s = v : spc (drop 1 r) where
    (v,r) = span (/= ':') s 

sendEmptyPuff h = do
    putWord32 h 0
    putWord32 h 8
    putWord32 h 0
    putWord32 h 0


emptyPuffer :: Handle -> ThreadId -> IO ()
emptyPuffer h _ = handle (\_ -> hClose h) $
    repeatM_ (threadDelay 7000000 >> sendEmptyPuff h)


connectThread :: Chan Puff -> [Category] -> [String] -> IO ()
connectThread c cs ds  = tryMost (trySeveral (map attemptConnect ds)) >>= \x -> case x of
    Left _ -> continue 
    Right h -> Control.Exception.handle (err h) $ do
	putWord32 h 1
	v <- readWord32 h
	--let ss = nub [c| (c,d') <- cs, d' == d]
	--    f c = catShowOld (c,d)
	--    gs = concatInter ":" (map f ss)
	--writeChan c (Message $ "Connected: " ++  d ++ " -- " ++ unwords ss)
	let gs = concatInter ":" (map catShowOld cs)
	putWord32 h 2
	putWord32 h (fromIntegral $ length gs * 2)
	putRaw h $ galeEncodeString gs
	mtid <- myThreadId
	stid <- forkIO (emptyPuffer h mtid)
	repeatM_ $ do
	    w <- readWord32 h
	    l <- readWord32 h
	    d <- readRaw h (fromIntegral l)
	    when (w == 0) $ do
		let (l,d') = xdrReadUInt d
		cat <- tryMapM parseCategoryOld (spc $ galeDecodeString (take (fromIntegral $ l ) d'))
		writeChan c $ Puff { signature = [], cats = cat, fragments = (decodeFragments $ drop (fromIntegral l + 4) d')} 
    where
	continue = threadDelay 2000000 >> connectThread c cs ds
	err h e@(AsyncException ThreadKilled) = do
	    hClose h
	    --writeChan c (Message $ "Disconnected: " ++ d)
	    throw e
	err h _ = do
	    hClose h
	    --writeChan c (Message $ "Disconnected: " ++ d)
	    continue
	
	
reconnectGaleContext gc = closeGaleContext gc >> connectGaleContext gc

closeGaleContext :: GaleContext -> IO ()
closeGaleContext gc = do
    thds <- takeMVar (threads gc)
    mapM_ killThread thds

connectGaleContext gc = do
    cats <- readMVar (categories gc)
    ps <- readMVar (proxy gc)
    thd <- forkIO $ connectThread (channel gc) cats (if null ps then nub (concatMap (hostStrings . snd) cats) else ps)
    putMVar (threads gc) [thd]

    
galeNextPuff :: GaleContext -> IO Puff
galeNextPuff gc = do
    p <- readChan $ channel gc
    p' <- galeDecryptPuff p
    return p'

galeDecryptPuff :: Puff -> IO Puff 
galeDecryptPuff p@(Puff {fragments = [("security/signature",FragmentData xs)]}) = tryElse p $ do 
    key <- parseKey $ take (fromIntegral (l - (8 + sb))) (drop (fromIntegral sb) xs'')
    galeDecryptPuff $ p {signature = (Unverifyable key): signature p, fragments = fl} where
	(l,xs') = xdrReadUInt xs
	(sb,xs'') = xdrReadUInt (drop 4 xs')
	fl = (decodeFragments $ drop (fromIntegral l + 4) xs')
galeDecryptPuff p@(Puff {fragments = [("security/encryption",FragmentData xs)]}) = tryElse p $ do
    (cd,ks) <- parseIO pe xs
    dfl <- first (map (td cd) [(x,y) | x@(_,n',_) <- ks, y@(Key n _,_) <- pks, n == n'])
    galeDecryptPuff p {signature = (Encrypted (map (\(_,n,_) -> n) ks)): signature p, fragments = dfl}  where
	pe = (parseExact cipher_magic1 >> pr parseNullString) <|> (parseExact cipher_magic2 >> pr parseLenString)
	pk pkname iv = do
	    kname <- pkname
	    keydata <- parseLenData
	    return $ (iv,kname,keydata)
	pr pkname = do
	    iv <- parseSome 8 
	    keycount <-  parseIntegral32 
	    ks <- replicateM keycount (pk pkname iv)
	    xs <- parseRest
	    return (xs,ks)
	td cd ((iv,kname,keydata),(_,rsa)) = do
	    dd <- decryptAll keydata iv rsa cd
	    --putLogLn $ show (map (chr. fromIntegral) dd)
	    let dfl = decodeFragments (drop 4 dd)
	    --putLogLn $ showFragments dfl
	    return dfl 
	pks = unsafePerformIO $ do
	    ks <- getPrivateKeys
	    fmap concat $ mapM (\x -> tryElse [] $ do
		rsa <- privkeyToRSA x
		pkey <- pkeyNewRSA rsa
		return [(x,pkey)]) ks
--		return (x,unsafePerformIO (privkeyToRSA x))) ks

--	    do rsa <- privkeyToRSA x return (x,rsa)) ks
--		(x,unsafePerformIO (privkeyToRSA x))) (unsafePerformIO getPrivateKeys)
galeDecryptPuff x = return x


showPuff (Puff {cats = s, fragments =  fl}) = unwords (map catShowNew s) ++ " \n" ++ concatMap (\(n,f) -> n ++ ": " ++ showFragment f ++ "\n") fl
--showPuff (Message s) = s

--showFragments fl = concatMap (\(n,f) -> n ++ ": " ++ showFragment f ++ "\n") fl

showFragment (FragmentText s) = s
showFragment (FragmentData xs) = "<DATA:" ++ show (length xs)  ++ ">"
showFragment (FragmentTime ct) = show ct
showFragment (FragmentInt x) = show x
showFragment (FragmentNest xs) = concatMap (\(n,f) -> n ++ ": " ++ showFragment f ++ "\n") xs

    

attemptConnect s = do
    h <- connectTo s (PortNumber galePort)
    hSetBuffering h NoBuffering
    return h



type XDRShowS = [Word8] -> [Word8]
type XDRReadS a = [Word8] -> (a,[Word8])



xdrWriteUInt :: Word32 -> XDRShowS
xdrWriteUInt x bs = (b1:b2:b3:b4:bs) where
    b1 = fromIntegral $ (x `shiftR` 24) .&. 0xFF
    b2 = fromIntegral $ (x `shiftR` 16) .&. 0xFF
    b3 = fromIntegral $ (x `shiftR` 8) .&. 0xFF
    b4 = fromIntegral $ x .&. 0xFF

xdrReadUInt :: XDRReadS Word32
xdrReadUInt (b1:b2:b3:b4:bs) = (x,bs) where
    x = (fromIntegral b4) .|. (fromIntegral b3 `shiftL` 8) .|. 
	    (fromIntegral b2 `shiftL` 16) .|. (fromIntegral b1 `shiftL` 24)
xdrReadUInt bs = error $ "xdrReadUInt " ++ show bs


putWord32 h x = do
    putRaw h $ xdrWriteUInt x []

readWord32 h = do
    xs <- readRaw h 4 
    return $ fst $ xdrReadUInt xs

galeEncodeString :: String -> [Word8]
galeEncodeString cs = concatMap (f . ord) cs where
    f x = (b1:b2:[]) where 
	b1 = fromIntegral $ (x `shiftR` 8) .&. 0xFF
	b2 = fromIntegral $ x .&. 0xFF

galeDecodeString :: [Word8] -> String
galeDecodeString [] = [] 
galeDecodeString (b1:b2:xs) = (chr $ (fromIntegral b2) .|. (fromIntegral b1 `shiftL` 8)) : galeDecodeString xs
galeDecodeString _ = error "invalid gale string"

---------------------
-- Gale stream coding
---------------------

parseWord32 :: GenParser Word8 Word32
parseWord32 = do
    [b1,b2,b3,b4] <- parseSome 4
    return $ (fromIntegral b4) .|. (fromIntegral b3 `shiftL` 8) .|. 
	    (fromIntegral b2 `shiftL` 16) .|. (fromIntegral b1 `shiftL` 24)

parseLenData = parseWord32 >>= parseSome

parseIntegral32 = fmap fromIntegral parseWord32

parseLenString :: GenParser Word8 String
parseLenString = do
    len <- parseIntegral32
    w <- parseSome (len * 2)
    return $ galeDecodeString w

parseNullString :: GenParser Word8 String
parseNullString = (sat (== 0) >> return "") <|> (parseSome 1 >>= \[x] -> fmap ((chr $ fromIntegral x):) parseNullString)
    

readNullTerminatedString :: [Word8] -> (String, [Word8])
readNullTerminatedString xs = f xs [] where
    f [] ret = (reverse (map (chr . fromIntegral) ret),[])
    f (0:xs) ret = (reverse (map (chr . fromIntegral) ret),xs)
    f (x:xs) ret = f xs (x:ret)

readLenString :: [Word8] -> (String, [Word8])
readLenString xs = (galeDecodeString (take l xs'), drop l xs') where
    (len,xs') = xdrReadUInt xs
    l = 2 * fromIntegral len

-------------------
-- initialization
-------------------

galeInit :: IO ()
galeInit =  do
    ks <- getPrivateKeys
    mapM_ (putStrLn . showKey) ks
    putStrLn "Done with keys."



-----------------
-- Key Management
-----------------


stons = map (fromIntegral . ord)

    


private_magic1 :: [Word8]
private_magic2 :: [Word8]
private_magic3 :: [Word8]
cipher_magic1 :: [Word8]
cipher_magic2 :: [Word8]
pubkey_magic1 :: [Word8]
pubkey_magic2 :: [Word8]
pubkey_magic3 :: [Word8]

cipher_magic1 = stons "h\DC3\002\000"
cipher_magic2 = stons "h\DC3\002\001"

private_magic1 = stons "h\DC3\0\001"
private_magic2 = stons "h\DC3\000\003"
private_magic3 = stons "GALE\000\002"


pubkey_magic1 = stons "h\DC3\000\000"
pubkey_magic2 = stons "h\DC3\000\002"
pubkey_magic3 = stons "GALE\000\001"

--key_response_cat = "_gale.key."
--key_request_cat = "_gale.query."

getRawContents :: String -> IO [Word8]
getRawContents fn = do
    c <- readFile fn
    return (map (fromIntegral . ord) c) 

getPrivateKeys :: IO [Key]
getPrivateKeys = do
    home <- getEnv "HOME"
    let p = (home ++ "/.gale/auth/private/")
    fs <- getDirectoryContents p
    ks <- mapM (\x -> tryMost (pk (p ++ x)) >>= \y -> return (x, y)) fs
    return [x| (_,Right x) <- ks] where
	pk n = getRawContents n >>= parseKey

keyDecode12 xs = fl where
    fl = [("rsa.modulus",FragmentData modulusD),
	  ("rsa.exponent",FragmentData exponentD),
	  ("rsa.private.exponent",FragmentData privateExponentD),
	  ("rsa.private.prime",FragmentData privatePrimeD),
	  ("rsa.private.prime.exponent",FragmentData privatePrimeExponentD),
	  ("rsa.private.coefficient",FragmentData privateCoefficientD),
	  ("rsa.bits", FragmentInt (fromIntegral bits))
	 ]
    (bits,ys) = xdrReadUInt xs
    (modulusD,xs') = splitRLE galeRSAModulusLen ys
    (exponentD,xs'') = splitRLE galeRSAModulusLen xs'
    (privateExponentD,xs''') = splitRLE (galeRSAPrimeLen * 2) xs''
    (privatePrimeD,xs'''') = splitRLE galeRSAModulusLen xs'''
    (privatePrimeExponentD,xs''''') = splitRLE (galeRSAPrimeLen * 2) xs''''
    (privateCoefficientD,_) = splitRLE galeRSAPrimeLen xs'''''
    
    

keyDecode3 xs = decodeFragments xs

parseKey :: [Word8] -> IO Key
parseKey xs | private_magic1 `isPrefixOf` xs = return $ Key n (keyDecode12 r)  where
    (n,r) = readNullTerminatedString (drop (length private_magic1) xs)
parseKey xs | private_magic2 `isPrefixOf` xs = return $ Key n (keyDecode12 r)  where
    (n,r) = readLenString (drop (length private_magic2) xs)
parseKey xs | (private_magic3) `isPrefixOf` xs = return $ Key n (keyDecode3 r)  where
    (n,r) = readLenString (drop (length private_magic3) xs)
parseKey xs = parseIO keyParse xs




showKey :: Key -> String
showKey (Key n fl) = "Key: " ++ n ++ "\n" ++ concatMap (\(n,f) -> n ++ ": " ++ showFragment f ++ "\n") fl


keyParse :: GenParser Word8 Key
keyParse = pk1 <|> pk2 <|> pk3 where
    pk1 = do
	parseExact pubkey_magic1
	kn <- parseNullString
	return $ Key kn []
	--(eof >> return (Key kn [])) <|> do
	--    undefined
    pk2 = do
	parseExact pubkey_magic2
	kn <- parseLenString
	return $ Key kn []
    pk3 = do
	parseExact pubkey_magic3
	kn <- parseLenString
	fl <- toParser decodeFragments 
	return $ Key kn fl



privkeyToRSA :: Key -> IO RSA
privkeyToRSA (Key _ fl)  = do
    rsa <- rsaNew
    fragmentData "rsa.modulus" fl >>= bn_bin2bn >>= rsaSetN rsa
    fragmentData "rsa.exponent" fl >>= bn_bin2bn >>= rsaSetE rsa
    fragmentData "rsa.private.exponent" fl >>= bn_bin2bn >>= rsaSetD rsa
    fragmentData "rsa.private.coefficient" fl >>= bn_bin2bn >>= rsaSetIQMP rsa
    xs <- fragmentData "rsa.private.prime" fl 
    let (p,q) = splitAt galeRSAPrimeLen xs
    bn_bin2bn p >>= rsaSetP rsa
    bn_bin2bn q >>= rsaSetQ rsa
    xs <- fragmentData "rsa.private.prime.exponent" fl 
    let (dmp1,dmq1) = splitAt galeRSAPrimeLen xs
    bn_bin2bn dmp1 >>= rsaSetDMP1 rsa
    bn_bin2bn dmq1 >>= rsaSetDMQ1 rsa
    rsaCheckKey rsa
    return rsa
    
---------------
-- data parsing
---------------

splitRLE :: Int -> [Word8] -> ([Word8], [Word8])
splitRLE n _ | n < 0 = error "invalid RLE encoding"
splitRLE 0 xs = ([],xs)
splitRLE n (c:xs) | c .&. 0x80 /= 0 = (take count xs ++ ys,rest) where
    count = fromIntegral $ (c .&. 0x7f) + 1
    (ys,rest) = splitRLE (n - count) (drop count xs) 
splitRLE n (c:x:xs) | c .&. 0x80 == 0 = (replicate count x ++ ys,rest) where
    count = fromIntegral $ (c .&. 0x7f) + 1
    (ys,rest) = splitRLE (n - count) xs 
splitRLE _ _  = error "invalid RLE encoding"

toParser :: ([c] -> a) -> GenParser c a
toParser f = MkP (\xs -> Just (f xs, []))

decodeFragments :: [Word8] -> [(String,Fragment)]
decodeFragments [] = []
decodeFragments xs = decodeFragment ft f : decodeFragments r where
    (ft,xs') = xdrReadUInt xs
    (fl,xs'') = xdrReadUInt xs'
    (f,r) = splitAt (fromIntegral fl) xs''

--decodeFragment ft f = (show ft ++ " " ++ show f,FragmentTime)
decodeFragment :: Word32 -> [Word8] -> (String, Fragment)
decodeFragment t xs = (fn,f t) where
    (fnl,xs') = xdrReadUInt xs
    (fn,xs'') = liftT2 (galeDecodeString,id) (splitAt (fromIntegral $ fnl * 2) xs')
    f 0 = FragmentText (galeDecodeString xs'')
    f 1 = FragmentData xs''
    f 2 = FragmentTime $ decodeTime xs''
    f 3 = FragmentInt (fromIntegral $ fst $ xdrReadUInt xs'')
    f 4 = FragmentNest (decodeFragments xs'')
    f x = error $ "unknown fragment: " ++ show x

decodeTime :: [Word8] -> ClockTime
decodeTime xs = TimeDiff {tdDay = 0, tdPicosec = 0, tdYear = 0, tdMonth = 0, tdHour = 0, tdMin = fromIntegral (t `div` 60) , tdSec = fromIntegral (t `mod` 60)} `addToClockTime` epoch where
    (t,_) = xdrReadUInt (drop 4 xs)

------------------
-- Examining puffs
------------------

fragmentData :: Monad m => String -> [(String,Fragment)] -> m [Word8]
fragmentData s fl = case lookup s fl of
    Just (FragmentData xs) -> return xs
    _ -> fail $ "fragment not found: " ++ s

getFragmentString :: Monad m => Puff -> String -> m String
getFragmentString (Puff {fragments = frags}) s = case lookup s frags of
    Just (FragmentText s) -> return s
    _ -> fail $ "fragment not found: " ++ s

getFragmentStrings :: Puff -> String -> [String]
getFragmentStrings (Puff {fragments = frags}) s = [v|(n,FragmentText v) <- frags, n == s]


