
-- $Id: GenUtil.hs,v 1.4 2002/08/29 06:06:26 john Exp john $


-- Copyright (c) 2002 John Meacham (john@foo.net)
-- 
-- Permission is hereby granted, free of charge, to any person obtaining a
-- copy of this software and associated documentation files (the
-- "Software"), to deal in the Software without restriction, including
-- without limitation the rights to use, copy, modify, merge, publish,
-- distribute, sublicense, and/or sell copies of the Software, and to
-- permit persons to whom the Software is furnished to do so, subject to
-- the following conditions:
-- 
-- The above copyright notice and this permission notice shall be included
-- in all copies or substantial portions of the Software.
-- 
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
-- OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
-- MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
-- IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
-- CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
-- TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
-- SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

----------------------------------------
-- utility functions written in pure H98
----------------------------------------

module GenUtil(
    putErr,putErrLn,putErrDie,concatInter,fromLeft,fromRight,repMaybe,
    liftT2, liftT3, liftT4, Unique(..), exitSuccess, exitFailure, epoch,
    fsts, snds, repeatM, repeatM_, replicateM, replicateM_, toMonad, 
    toMonadM, splitEither, shellQuote, simpleQuote, simpleUnquote
    ) where

import System
import qualified IO
import List(intersperse)
import Time
import Monad
import Char(isAlphaNum, isSpace)

putErr = IO.hPutStr IO.stderr
putErrLn s = putErr (s ++ "\n")

putErrDie s = putErrLn s >> System.exitFailure

concatInter x = concat . (intersperse x)

exitSuccess = System.exitWith System.ExitSuccess

instance Show (a -> b) where
    show x = "<function>"


fromRight (Right x) = x
fromRight _ = error "fromRight"

fromLeft (Left x) = x
fromLeft _ = error "fromLeft"

repMaybe :: (a -> Maybe a) -> a -> a
repMaybe f e = case f e of 
    Just e' -> repMaybe f e'
    Nothing -> e

liftT4 (f1,f2,f3,f4) (v1,v2,v3,v4) = (f1 v1, f2 v2, f3 v3, f4 v4)
liftT3 (f,g,h) (x,y,z) = (f x, g y, h z)
liftT2 (f,g) (x,y) = (f x, g y)


-- class for monads which can generate
-- unique values
class Monad m => Unique m where
    newUniq :: m Int
    peekUniq :: m Int
    modifyUniq :: (Int -> Int) -> m ()
    newUniq = do
	v <- peekUniq
	modifyUniq (+1) 
	return v



epoch = toClockTime $ CalendarTime { ctYear = 1970, ctMonth = January, ctDay = 0, ctHour = 0, ctMin = 0, ctSec = 0, ctTZ = 0, ctPicosec = 0, ctWDay = undefined, ctYDay = undefined, ctTZName = undefined, ctIsDST = undefined}

fsts = map fst
snds = map snd

repeatM x = sequence $ repeat x
repeatM_ x = sequence_ $ repeat x

replicateM n x = sequence $ replicate n x
replicateM_ n x = sequence_ $ replicate n x

toMonad :: Monad m => Maybe a -> m a
toMonad (Just x) = return x
toMonad Nothing = fail "Nothing"

toMonadM :: Monad m => m (Maybe a) -> m a
toMonadM action = join $ liftM toMonad action

splitEither :: [Either a b] -> ([a],[b])
splitEither ((Left x):r) = let (xs,ys) = splitEither r in (x:xs,ys)
splitEither ((Right y):r) = let (xs,ys) = splitEither r in (xs,y:ys)
splitEither [] = ([],[])


simpleQuote :: [String] -> String
simpleQuote ss = unwords (map f ss) where
    f s | any isBad s = "'" ++ dquote s ++ "'"
    f s = s
    dquote s = concatMap (\c -> if c == '\'' then "''" else [c]) s
    isBad c = isSpace c || c == '\''

simpleUnquote :: String -> [String]
simpleUnquote s = f (dropWhile isSpace s)  where
    f [] = []
    f ('\'':xs) = let (x,y) = quote' "" xs in x:f (dropWhile isSpace y)
    f xs = let (x,y) = span (not . isSpace) xs in x:f (dropWhile isSpace y)
    quote' a ('\'':'\'':xs) = quote' ('\'':a) xs
    quote' a ('\'':xs) = (reverse a, xs)
    quote' a (x:xs) = quote' (x:a) xs
    quote' a [] = (reverse a, "")

shellQuote :: [String] -> String
shellQuote ss = unwords (map f ss) where
    f s | any (not . isGood) s = "'" ++ dquote s ++ "'"
    f s = s
    dquote s = concatMap (\c -> if c == '\'' then "'\\''" else [c]) s
    isGood c = isAlphaNum c || c `elem` "@/."



    

{-
    
shellQuote :: [String] -> String
shellQuote ss = unwords (map (\s -> if all isGood s then s else "'" ++ f s ++ "'") ss) where
    f "" = ""
    f ('\\':xs) = "\\\\" ++ f xs
    f ('\'':xs) = "\\'" ++ f xs
    f (x:xs) = x : f xs
    isGood c = isAlphaNum c || c `elem` "@/."



shellUnquote :: String -> [String]
shellUnquote s = map uq (words s) where --TODO
    uq ('\'':xs) = init xs



-}
