module Puff(Puff(..),Fragment(..),Signature(..),Key(..),Category, readPuffs, writePuffs, emptyPuff,
	    catParseNew, catShowNew, parseCategoryOld, catShowOld) where

import Word
import Int
import Time
import Bin
import Error
import List
import SimpleParser

type Category = (String,String)

data Puff = Puff {cats :: [Category], signature :: [Signature], fragments :: [(String,Fragment)]} 
data Fragment = FragmentText String | FragmentData [Word8] | FragmentTime ClockTime | FragmentInt Int32 | FragmentNest [(String,Fragment)]
data Signature = Unverifyable Key | Signed Key | Encrypted [String]
data Key = Key String [(String,Fragment)]

---------------------
-- working with puffs
---------------------

emptyPuff = Puff {cats = [], signature = [], fragments = []}

catfixes = [ ("/", ".|"), (".", "/"), (":", "..") ]

    
catParseNew :: String -> Category
catParseNew cs = (c,(drop 1) d) where
    (c,d) = span (/= '@') cs
    
parseCategoryOld :: Monad m => String -> m Category
parseCategoryOld = parser p where
    con cs | [nv] <- [x ++ (con $ drop (length y) cs) |(x,y) <- catfixes, y `isPrefixOf` cs] = nv
    con (c:cs) = c:con cs
    con "" = "" 
    bl [] = []
    bl [x] = []
    bl (x:xs) = x:bl xs
    p = do
	char '@'
	d <- many (noneOf "/")
	parseExact "/user/"
	c <- parseRest
	return (con (bl c),d)

catShowOld :: Category -> String
catShowOld (c,d) = "@" ++ d ++ "/user/" ++ con c where 
    con cs | [nv] <- [x ++ (con $ drop (length y) cs) |(y,x) <- catfixes, y `isPrefixOf` cs] = nv
    con (c:cs) = c:con cs
    con "" = "" 

catShowNew :: Category -> String
catShowNew (x,"") = x 
catShowNew (x,y) = x ++ "@" ++ y

----------------
-- Puff Pickling
----------------

writePuffs :: String -> [Puff] -> IO ()
writePuffs fn ps = attempt $ putFile fn ps

readPuffs :: String -> IO [Puff]
readPuffs fn = tryElse [] $ getFile fn


---------------------
-- generated by DrIFT
---------------------

instance Binary Puff where
    put bh (Puff aa ab ac) = do
	    pos <- putBits bh 1 0
	    put bh aa
	    put bh ab
	    put bh ac
	    return pos
    get bh = do
	    h <- getBits bh 1
	    case h of
	      0 -> do
		    aa <- get bh
		    ab <- get bh
		    ac <- get bh
		    return (Puff aa ab ac)
    sizeOf (Puff aa ab ac) = 1 + sizeOf aa + sizeOf ab + sizeOf ac

instance Binary Fragment where
    put bh (FragmentText aa) = do
	    pos <- putBits bh 3 0
	    put bh aa
	    return pos
    put bh (FragmentData ab) = do
	    pos <- putBits bh 3 1
	    put bh ab
	    return pos
    put bh (FragmentTime ac) = do
	    pos <- putBits bh 3 2
	    put bh ac
	    return pos
    put bh (FragmentInt ad) = do
	    pos <- putBits bh 3 3
	    put bh ad
	    return pos
    put bh (FragmentNest ae) = do
	    pos <- putBits bh 3 4
	    put bh ae
	    return pos
    get bh = do
	    h <- getBits bh 3
	    case h of
	      0 -> do
		    aa <- get bh
		    return (FragmentText aa)
	      1 -> do
		    ab <- get bh
		    return (FragmentData ab)
	      2 -> do
		    ac <- get bh
		    return (FragmentTime ac)
	      3 -> do
		    ad <- get bh
		    return (FragmentInt ad)
	      4 -> do
		    ae <- get bh
		    return (FragmentNest ae)
    sizeOf (FragmentText aa) = 3 + sizeOf aa
    sizeOf (FragmentData ab) = 3 + sizeOf ab
    sizeOf (FragmentTime ac) = 3 + sizeOf ac
    sizeOf (FragmentInt ad) = 3 + sizeOf ad
    sizeOf (FragmentNest ae) = 3 + sizeOf ae

instance Binary Signature where
    put bh (Unverifyable aa) = do
	    pos <- putBits bh 2 0
	    put bh aa
	    return pos
    put bh (Signed ab) = do
	    pos <- putBits bh 2 1
	    put bh ab
	    return pos
    put bh (Encrypted ac) = do
	    pos <- putBits bh 2 2
	    put bh ac
	    return pos
    get bh = do
	    h <- getBits bh 2
	    case h of
	      0 -> do
		    aa <- get bh
		    return (Unverifyable aa)
	      1 -> do
		    ab <- get bh
		    return (Signed ab)
	      2 -> do
		    ac <- get bh
		    return (Encrypted ac)
    sizeOf (Unverifyable aa) = 2 + sizeOf aa
    sizeOf (Signed ab) = 2 + sizeOf ab
    sizeOf (Encrypted ac) = 2 + sizeOf ac

instance Binary Key where
    put bh (Key aa ab) = do
	    pos <- putBits bh 0 0
	    put bh aa
	    put bh ab
	    return pos
    get bh = do
	    h <- getBits bh 0
	    case h of
	      0 -> do
		    aa <- get bh
		    ab <- get bh
		    return (Key aa ab)
    sizeOf (Key aa ab) = 0 + sizeOf aa + sizeOf ab

--  Imported from other files :-
