{-# OPTIONS -#include "RSA_hsc.h" #-}
{-# LINE 1 "RSA.hsc" #-}
module RSA where
{-# LINE 2 "RSA.hsc" #-}



{-# LINE 5 "RSA.hsc" #-}

{-# LINE 6 "RSA.hsc" #-}

{-# LINE 7 "RSA.hsc" #-}

import Foreign
import CForeign
import MarshalArray
--import GHC.Weak
import Control.Exception as E
import IOExts

data RSAStruct 
data EVPCipherCtx
data EVPCipher
data EVP_PKEY
data EVP_CIPHER
data EVP_CIPHER_CTX
data BIGNUM

type RSA = Ptr RSAStruct
type EVPCIPHERCTX = Ptr EVPCipherCtx
type EVPCIPHER = Ptr EVPCipher

------------------------
-- marshalling utilities
------------------------

throwZero_ s = throwIf_ (== 0) (const s)

withData :: [Word8] -> (Ptr CUChar -> CInt -> IO a) -> IO a
withData xs f = withArray xs (\a -> f (castPtr a) (fromIntegral $ length xs))

returnData :: Int -> (Ptr CUChar -> Ptr CInt -> IO ()) -> IO [Word8]
returnData sz f = do
	alloca (\bp -> allocaArray sz (\m -> f m bp >> peek bp >>= \s -> fmap (map fromIntegral) (peekArray (fromIntegral s) m)))

---------------
-- RSA routines
---------------


foreign import ccall unsafe "RSA_new" rsaNew :: IO RSA
foreign import ccall unsafe "RSA_free" rsaFree :: RSA -> IO ()

foreign import ccall unsafe "RSA_check_key" rsa_check_key :: RSA -> IO CInt

rsaCheckKey :: RSA -> IO ()
rsaCheckKey rsa = throwZero_ "RSA_check_key" $ rsa_check_key rsa


rsaSetN :: RSA -> Ptr BIGNUM -> IO ()
rsaSetN = ((\hsc_ptr -> pokeByteOff hsc_ptr 12))
{-# LINE 56 "RSA.hsc" #-}

rsaSetE :: RSA -> Ptr BIGNUM -> IO ()
rsaSetE = ((\hsc_ptr -> pokeByteOff hsc_ptr 16))
{-# LINE 59 "RSA.hsc" #-}

rsaSetD :: RSA -> Ptr BIGNUM -> IO ()
rsaSetD = ((\hsc_ptr -> pokeByteOff hsc_ptr 20))
{-# LINE 62 "RSA.hsc" #-}

rsaSetP :: RSA -> Ptr BIGNUM -> IO ()
rsaSetP = ((\hsc_ptr -> pokeByteOff hsc_ptr 24))
{-# LINE 65 "RSA.hsc" #-}

rsaSetQ :: RSA -> Ptr BIGNUM -> IO ()
rsaSetQ = ((\hsc_ptr -> pokeByteOff hsc_ptr 28))
{-# LINE 68 "RSA.hsc" #-}

rsaSetDMP1 :: RSA -> Ptr BIGNUM -> IO ()
rsaSetDMP1 = ((\hsc_ptr -> pokeByteOff hsc_ptr 32))
{-# LINE 71 "RSA.hsc" #-}

rsaSetDMQ1 :: RSA -> Ptr BIGNUM -> IO ()
rsaSetDMQ1 = ((\hsc_ptr -> pokeByteOff hsc_ptr 36))
{-# LINE 74 "RSA.hsc" #-}

rsaSetIQMP :: RSA -> Ptr BIGNUM -> IO ()
rsaSetIQMP = ((\hsc_ptr -> pokeByteOff hsc_ptr 40))
{-# LINE 77 "RSA.hsc" #-}




----------------------
-- Envelope Encryption 
----------------------

foreign import ccall unsafe "EVP_OpenInit" evpOpenInit :: Ptr EVP_CIPHER_CTX -> Ptr EVP_CIPHER -> Ptr CUChar -> CInt -> Ptr CUChar -> Ptr EVP_PKEY -> IO CInt
foreign import ccall unsafe "EVP_DecryptUpdate" evpOpenUpdate :: Ptr EVP_CIPHER_CTX -> Ptr CUChar -> Ptr CInt -> Ptr CUChar -> CInt -> IO CInt 
foreign import ccall unsafe "EVP_OpenFinal" evpOpenFinal :: Ptr EVP_CIPHER_CTX -> Ptr CUChar -> Ptr CInt -> IO Int 

foreign import ccall unsafe "EVP_des_ede3_cbc" evpDesEde3Cbc :: IO (Ptr EVP_CIPHER)


foreign import ccall unsafe evpCipherContextBlockSize :: Ptr EVP_CIPHER_CTX -> IO Int 

{-# LINE 94 "RSA.hsc" #-}


evp_OpenInit :: Ptr EVP_CIPHER_CTX -> Ptr EVP_CIPHER -> [Word8] -> [Word8] -> Ptr EVP_PKEY -> IO ()
evp_OpenInit cctx cipher ek iv pkey =  do
	throwZero_ "EVP_OpenInit" $ withData ek (\a b -> withArray iv (\iv -> evpOpenInit cctx cipher a b (castPtr iv) pkey))
	return ()
	
evp_OpenUpdate :: Ptr EVP_CIPHER_CTX -> [Word8] -> IO [Word8]
evp_OpenUpdate cctx ind = do
    bsz <- evpCipherContextBlockSize cctx
    let sz = length ind + bsz
    withData ind (\ina inl -> returnData sz (\outa outl -> throwZero_ "EVP_OpenUpdate" (evpOpenUpdate cctx outa outl ina inl)))

evp_OpenFinal ::  Ptr EVP_CIPHER_CTX -> IO [Word8]
evp_OpenFinal cctx = do
    bsz <- evpCipherContextBlockSize cctx
    d <- returnData bsz (\outa outl -> throwZero_ "EVP_OpenFinal" (evpOpenFinal cctx outa outl))
    return d
    
    

foreign import ccall unsafe "EVP_PKEY_new" evpPKEYNew :: IO (Ptr EVP_PKEY)
foreign import ccall unsafe "EVP_PKEY_assign" evpPKEYAssign :: Ptr EVP_PKEY -> CInt -> Ptr a -> IO CInt 
foreign import ccall unsafe "EVP_PKEY_free" evpPKEYFree :: Ptr EVP_PKEY -> IO ()

foreign import ccall unsafe "EVP_CIPHER_CTX_init" evpCipherCtxInit :: Ptr EVP_CIPHER_CTX -> IO CInt
foreign import ccall unsafe "EVP_CIPHER_CTX_cleanup" evpCipherCtxCleanup :: Ptr EVP_CIPHER_CTX -> IO CInt

withCipherCtx :: (Ptr EVP_CIPHER_CTX -> IO a) -> IO a
withCipherCtx action = allocaBytes (4216) $ \cctx -> 
{-# LINE 124 "RSA.hsc" #-}
	    E.bracket_ (throwZero_ "EVP_CIPHER_CTX_init" $ evpCipherCtxInit cctx) 
		(throwZero_ "EVP_CIPHER_CTX_cleanup" $ evpCipherCtxCleanup cctx) 
		    (action cctx)
    
foreign import ccall unsafe pkeyNewRSA :: RSA -> IO (Ptr EVP_PKEY)

{-# LINE 130 "RSA.hsc" #-}

decryptAll :: [Word8] -> [Word8] -> Ptr EVP_PKEY -> [Word8] -> IO [Word8]
decryptAll keydata iv pkey xs = withCipherCtx $ \cctx -> do
	    d <- withData keydata $ \a b -> withArray iv $ \iv -> do
		throwZero_ "EVP_OpenInit" $ evpOpenInit cctx cipher a b (castPtr iv) pkey
		d <- evp_OpenUpdate cctx xs
		dr <- evp_OpenFinal cctx
		return $ d ++ dr  
	    return d where
	cipher = unsafePerformIO evpDesEde3Cbc



--------------------
-- BigNum routines
--------------------

bn_bin2bn :: [Word8] -> IO (Ptr BIGNUM)
bn_bin2bn xs = throwIfNull "BN_bin2bn" $ withData xs (\a b -> bnBin2Bn a b nullPtr)

foreign import ccall unsafe "BN_bin2bn" bnBin2Bn :: Ptr CUChar -> CInt -> Ptr BIGNUM -> IO (Ptr BIGNUM)




------------
------------
------------

{- graveyard
foreign import ccall unsafe "BN_bn2bin" bnBn2Bin :: Ptr BIGNUM -> Ptr CUChar -> IO CInt


-- int RSA_public_encrypt(int flen, unsigned char *from, unsigned char *to, RSA *rsa, int padding);
foreign import ccall unsafe "RSA_public_encrypt" rsaPublicEncrypt :: CInt -> Ptr CUChar -> Ptr CUChar -> RSA -> CInt -> IO CInt

-- int RSA_private_decrypt(int flen, unsigned char *from, unsigned char *to, RSA *rsa, int padding);
foreign import ccall unsafe "RSA_public_decrypt" rsaPublicDecrypt :: CInt -> Ptr CUChar -> Ptr CUChar -> RSA -> CInt -> IO CInt

--int RSA_sign(int type, unsigned char *m, unsigned int m_len, unsigned char *sigret, unsigned int *siglen, RSA *rsa);
foreign import ccall unsafe "RSA_sign" rsaSign :: CInt -> Ptr CUChar -> CUInt -> Ptr CUChar -> Ptr CUInt -> RSA -> IO CInt

--int RSA_size(RSA *rsa);
foreign import ccall unsafe "RSA_size" rsaSize :: RSA -> IO CInt

-- int RSA_verify(int type, unsigned char *m, unsigned int m_len, unsigned char *sigbuf, unsigned int siglen, RSA *rsa);
foreign import ccall unsafe "RSA_verify" rsaVerify :: CInt -> Ptr CUChar -> CUInt -> Ptr CUChar -> CUInt -> RSA -> IO CInt

-- RSA *RSA_generate_key(int num, unsigned long e, void (*callback)(int,int,void *), void *cb_arg);
foreign import ccall "RSA_generate_key" rsaGenerateKey :: CInt -> CULong -> (FunPtr (CInt -> CInt -> Ptr () -> IO ())) -> Ptr () -> IO RSA

-- int RSA_check_key(RSA *rsa);
foreign import ccall unsafe "RSA_check_key" rsaCheckKey :: RSA -> IO CInt

-}
{-

decryptAll :: [Word8] -> [Word8] -> RSA -> [Word8] -> IO [Word8]
decryptAll keydata iv rsa xs = do
	    pkey <- pkeyNewRSA rsa
	    cctx <- evp_OpenInit cipher keydata iv pkey
	    d <- evp_OpenUpdate cctx xs
	    dr <- evp_OpenFinal cctx
	    evpPKEYFree pkey
	    return $ d ++ dr where 
		cipher = unsafePerformIO  evpDesEde3Cbc
    
  -}  


